from __future__ import unicode_literals

import errno
import functools
import mock
import os
import pathlib
import shutil
import subprocess
import tempfile
import unittest
import re

from six import BytesIO

import zc.buildout

import rubygems


def touch(path):
    with path.open('w') as f:
        f.write('')

gem_dependencies = {
    'cucumber-core': {
        '3.0.0': [
            {
                'gemname': 'backports',
                'constraint_list': [
                    {
                        'symbol': '>=',
                        'version': '3.8.0',
                    },
                    {
                        'symbol': '',
                        'version': '',
                    },
                ]
            },
            {
                'gemname': 'cucumber-tag_expressions',
                'constraint_list': [
                    {
                        'symbol': '>=',
                        'version': '1.0.1',
                    },
                    {
                        'symbol': '',
                        'version': '',
                    },
                ]
            },
            {
                'gemname': 'gherkin',
                'constraint_list': [
                    {
                        'symbol': '>=',
                        'version': '4.1.3',
                    },
                    {
                        'symbol': '',
                        'version': '',
                    },
                ]
            },
        ],
        '7.1.0': [
            {
                'gemname': 'cucumber-gherkin',
                'constraint_list': [
                    {
                        'symbol': '~>',
                        'version': '14.0',
                    },
                    {
                        'symbol': '>=',
                        'version': '14.0.1',
                    },
                ]
            },
            {
                'gemname': 'cucumber-messages',
                'constraint_list': [
                    {
                        'symbol': '>=',
                        'version': '12.2.0',
                    },
                    {
                        'symbol': '~>',
                        'version': '12.2',
                    },
                ]
            },
            {
                'gemname': 'cucumber-tag_expressions',
                'constraint_list': [
                    {
                        'symbol': '~>',
                        'version': '2.0',
                    },
                    {
                        'symbol': '>=',
                        'version': '2.0.4',
                    },
                ]
            },
        ],
    },
    'cucumber-gherkin': {
        '9.1.0': [
            {
                'gemname': 'cucumber-messages',
                'constraint_list': [
                    {
                        'symbol': '~>',
                        'version': '9.0',
                    },
                    {
                        'symbol': '>=',
                        'version': '9.0.3',
                    },
                ]
            },
        ],
        '14.1.0': [
            {
                'gemname': 'cucumber-messages',
                'constraint_list': [
                    {
                        'symbol': '~>',
                        'version': '12.3',
                    },
                    {
                        'symbol': '>=',
                        'version': '12.3.2',
                    },
                ]
            },
        ],
    },
    'gherkin': {
        '8.0.0': [
            {
                'gemname': 'cucumber-messages',
                'constraint_list': [
                    {
                        'symbol': '~>',
                        'version': '6.0',
                    },
                    {
                        'symbol': '>=',
                        'version': '6.0.1',
                    },
                ]
            },
        ],
    },    
    'hoe': {
        '3.4.0': [
            {
                'gemname': 'rake',
                'constraint_list': [
                    {
                        'symbol': '<',
                        'version': '11.0',
                    },
                    {
                        'symbol': '>=',
                        'version': '0.8',
                    },
                ]
            },
        ],
    },
    'rash': {
        '0.1.0': [
            {
                'gemname': 'hashie',
                'constraint_list': [
                    {
                        'symbol': '>=',
                        'version': '0.3.1',
                    },
                    {
                        'symbol': '',
                        'version': '',
                    },
                ]
            },
        ],
        '0.3.0': [
            {
                'gemname': 'hashie',
                'constraint_list': [
                    {
                        'symbol': '~>',
                        'version': '1.0.0',
                    },
                    {
                        'symbol': '',
                        'version': '',
                    },
                ]
            },
        ],
    },
}


class fixture(object):
    def __init__(self, options=None, version='2.0.0'):
        self.options = options or {}
        self.version = version

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(test):
            buildout, name, options = self.set_up()
            cwd = os.getcwd()
            os.chdir(str(self.tempdir))
            func(test, self.tempdir, self.patches, buildout, name, options)
            os.chdir(cwd)
            self.tear_down()
        return wrapper

    def patch(self, modules):
        self.patchers = {}
        self.patches = {}
        for name, module in modules:
            self.patchers[name] = mock.patch(module)
            self.patches[name] = self.patchers[name].start()

    def makedirs(self, dirs):
        self.tempdir = pathlib.Path(tempfile.mkdtemp())
        for directory in dirs:
            os.makedirs(str(self.tempdir / directory))

    def set_up(self):
        name = 'rubygems'
        version = self.options.get('return', {}).get('version', self.version)
 
        self.patch((
            ('check_output', 'rubygems.subprocess.check_output'),
            ('urlopen', 'rubygems.urllib.request.urlopen'),
            ('download', 'rubygems.Download'),
        ))
        self.patches['urlopen'].return_value = BytesIO(
            b'https://rubygems.org/rubygems/rubygems-2.0.0.zip'
        )

        self.makedirs((
            'bin',
            'ruby-' + version,
            'rubygems-' + version,
            'rubygems/bin',
        ))

        buildout = {'buildout': dict({
            'parts-directory': str(self.tempdir),
            'bin-directory': str(self.tempdir / 'bin'),
        }, **self.options.get('buildout', {}))}

        options = self.options.get('recipe', {})

        return buildout, name, options

    def tear_down(self):
        for patcher in self.patchers.values():
            patcher.stop()
        shutil.rmtree(str(self.tempdir))


class RubyGemsTests(unittest.TestCase):
    def check_output_test(self, check_output_mock, expected_arg_list_list):
        self.assertEqual(check_output_mock.call_count,
                          len(expected_arg_list_list))

        for command_nb, expected_arg_list in enumerate(
                                                expected_arg_list_list):
            # half of the mock calls come from the use of 'decode' method
            self.assertEqual(
                check_output_mock.mock_calls[2*command_nb][1][0],
                expected_arg_list
            )
            self.assertEqual(
                str(check_output_mock.mock_calls[2*command_nb+1]),
                'call().decode()'
            )

    def install_with_default_rubygems_test(self, path, patches,
                                            expected_install_arg_list_list):
        # One urlopen call to get latest version
        self.assertEqual(patches['urlopen'].call_count, 1)

        args = patches['urlopen'].mock_calls[0][1]
        self.assertEqual(args, ('https://rubygems.org/pages/download',))

        # One slapos.recipe.downloadunpacked call to download rubygems
        self.assertEqual(patches['download'].call_count, 1)

        args = patches['download'].mock_calls[0][1]
        self.assertEqual(args[2], {
            'url': 'https://rubygems.org/rubygems/rubygems-2.0.0.zip',
            'destination': str(path / 'rubygems-2.0.0'),
        })

        expected_install_arg_list_list.insert(0, [
            'ruby', 'setup.rb', 'all', '--prefix=%s/rubygems' % path,
            '--no-rdoc', '--no-ri',
        ])
        self.check_output_test(
            patches['check_output'], expected_install_arg_list_list)

    @fixture({'recipe': {'gems': 'sass'}})
    def test_success(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
            [
                'ruby', None, 'install', '--no-document',
                '--bindir=%s/rubygems/bin' % path,
                'sass', '--',
            ],
        ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list) 

    @fixture({'recipe': {}})
    def test_missing_gems(self, path, patches, buildout, name, options):
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            "Configuration error, 'gems' option is missing",
            rubygems.Recipe, buildout, name, options
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_oserror(self, path, patches, buildout, name, options):
        patches['check_output'].side_effect = OSError
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'System error, command failed: .*',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_signal_received(self, path, patches, buildout, name, options):
        exception = subprocess.CalledProcessError(-1, '')
        patches['check_output'].side_effect = exception
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'System error, command received signal 1: .*',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_non_zero_exitcode(self, path, patches, buildout, name, options):
        exception = subprocess.CalledProcessError(1, '')
        patches['check_output'].side_effect = exception
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'System error, command failed with exit code 1: .*',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_update(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.update()

    @fixture({'recipe': {'gems': 'sass', 'environment': 'invalid'}})
    def test_invalid_environment(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
          zc.buildout.UserError,
          'Configuration error, '
          'every environment line should contain a "=" sign',
          recipe.install
        )

    @fixture({'recipe': {
        'gems': 'sass',
        'url': 'https://rubygems.org/rubygems/rubygems-2.0.0.zip',
    }})
    def test_version_from_url(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

    @fixture({'recipe': {'gems': 'sass', 'version': '2.0.0'}})
    def test_version(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

    @fixture({'recipe': {'gems': 'sass'}})
    def test_no_version(self, path, patches, buildout, name, options):
        patches['urlopen'].return_value = BytesIO(b'')
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(zc.buildout.UserError,
            'Can\'t find latest rubygems version.',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    @mock.patch('rubygems.os.mkdir')
    def test_mkdir_error(self, path, patches, buildout, name, options, mkdir):
        mkdir.side_effect = OSError(errno.EIO)
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'IO error while creating %s/rubygems directory.' % path,
            recipe.install
        )

    @fixture({'recipe': {'gems': 'sass'}})
    def test_executables(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        touch(pathlib.Path(recipe.options['location']) / 'bin/sass')
        recipe.install()

    @fixture({'recipe': {'gems': 'sass==1.0'}})
    def test_pinned_versions(self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        touch(path / 'rubygems/bin/gem')
        recipe.install()

        args = patches['check_output'].mock_calls[0][1]
        for arg in args[0]:
            matched = re.search('^--version.*$', arg)
            if matched:
                self.assertEqual(matched.group(), '--version=1.0')
                break

    @fixture({'recipe': {'gems': 'sass', 'deployment': 'true'}})
    def test_deployment_not_pinned_version_error(
            self, path, patches, buildout, name, options):
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version for gem sass is missing',
            recipe.install
        )

    @fixture({'recipe': {'gems': 'rash==0.1.0', 'deployment': 'true'}})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_not_pinned_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [gem_dependencies['rash']['0.1.0']]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version for dependency hashie is missing',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'rash==0.1.0 hashie==0.3.0',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_rash010_version_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [gem_dependencies['rash']['0.1.0']]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 0.3.0 for gem hashie '
            'does not satisfy dependency constraint >= 0.3.1 of gem rash',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'rash==0.1.0 hashie==0.3.1',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_rash010_pinned_dependency(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [gem_dependencies['rash']['0.1.0'],
                                            [],]
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'rash', '--version=0.1.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'hashie', '--version=0.3.1', '--',
              ],
          ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)

    @fixture({'recipe': {
        'gems': 'rash==0.3.0 hashie==1.1.0',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_rash030_version_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [gem_dependencies['rash']['0.3.0']]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 1.1.0 for gem hashie '
            'does not satisfy dependency constraint < 1.1 of gem rash',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'rash==0.3.0 hashie==1.0.0',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_rash030_pinned_dependency(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [gem_dependencies['rash']['0.3.0'],
                                            [],]
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'rash', '--version=0.3.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'hashie', '--version=1.0.0', '--',
              ],
          ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)

    @fixture({'recipe': {
        'gems': 'cucumber-gherkin==9.1.0 cucumber-messages==9.0.2',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_cucumber_gherkin_version_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['cucumber-gherkin']['9.1.0']]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 9.0.2 for gem cucumber-messages '
            'does not satisfy dependency constraint >= 9.0.3 '
            'of gem cucumber-gherkin',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'cucumber-gherkin==9.1.0 cucumber-messages==9.0.3',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_cucumber_gherkin_pinned_dependency(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['cucumber-gherkin']['9.1.0'], []]
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-gherkin',
                  '--version=9.1.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-messages',
                  '--version=9.0.3', '--',
              ],
          ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)

    @fixture({'recipe': {
        'gems': 'hoe==3.4.0 rake==11.0.0',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_hoe_version_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [gem_dependencies['hoe']['3.4.0']]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 11.0.0 for gem rake '
            'does not satisfy dependency constraint < 11.0 of gem hoe',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'hoe==3.4.0 rake==10.0.1',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_hoe_pinned_dependency(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['hoe']['3.4.0'], []]
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'hoe', '--version=3.4.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'rake', '--version=10.0.1', '--',
              ],
          ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)

    @fixture({'recipe': {
        'gems': 'cucumber-core==3.0.0 backports==3.8.0 '
                'cucumber-tag_expressions==1.0.1 gherkin==8.0.0',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_cucumber_core300_version_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['cucumber-core']['3.0.0'], [], [],
            gem_dependencies['gherkin']['8.0.0'],
        ]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version for dependency '
            'cucumber-messages is missing',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'cucumber-core==3.0.0 backports==3.8.0 '
                'cucumber-tag_expressions==1.0.1 gherkin==8.0.0 '
                'cucumber-messages==6.0.1',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_cucumber_core300_pinnned_dependency(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['cucumber-core']['3.0.0'], [], [],
            gem_dependencies['gherkin']['8.0.0'], [],
        ]
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-core',
                  '--version=3.0.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'backports',
                  '--version=3.8.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-tag_expressions',
                  '--version=1.0.1', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'gherkin',
                  '--version=8.0.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-messages',
                  '--version=6.0.1', '--',
              ],
          ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)

    @fixture({'recipe': {
        'gems': 'cucumber-core==7.1.0 cucumber-messages==12.2.0 '
                'cucumber-gherkin==14.1.0 cucumber-tag_expressions==2.0.4',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_cucumber_core710_version_dependency_error(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['cucumber-core']['7.1.0'], [],
            gem_dependencies['cucumber-gherkin']['14.1.0'], [],
        ]
        recipe = rubygems.Recipe(buildout, name, options)
        self.assertRaisesRegexp(
            zc.buildout.UserError,
            'Configuration error, version 12.2.0 for gem cucumber-messages '
            'does not satisfy dependency constraint >= 12.3 '
            'of gem cucumber-gherkin',
            recipe.install
        )

    @fixture({'recipe': {
        'gems': 'cucumber-core==7.1.0 cucumber-messages==12.3.2 '
                'cucumber-gherkin==14.1.0 cucumber-tag_expressions==2.0.4',
        'deployment': 'true'
    }})
    @mock.patch('rubygems.Recipe.get_dependency_list')
    def test_deployment_cucumber_core710_pinned_dependency(
            self, path, patches, buildout, name, options, get_dependency_list):
        get_dependency_list.side_effect = [
            gem_dependencies['cucumber-core']['7.1.0'], [],
            gem_dependencies['cucumber-gherkin']['14.1.0'], [],
        ]
        recipe = rubygems.Recipe(buildout, name, options)
        recipe.install()

        expected_install_arg_list_list = [
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-core',
                  '--version=7.1.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-messages',
                  '--version=12.3.2', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-gherkin',
                  '--version=14.1.0', '--',
              ],
              [
                  'ruby', None, 'install', '--no-document',
                  '--bindir=%s/rubygems/bin' % path,
                  '--ignore-dependencies', 'cucumber-tag_expressions',
                  '--version=2.0.4', '--',
              ],
          ]
        self.install_with_default_rubygems_test(
            path, patches, expected_install_arg_list_list)